unit DBCoords;
{$S-,W-,R-,F+}
{$C PRELOAD}

interface

uses WinTypes, WinProcs, Messages, SysUtils, Classes, Controls, Forms, Menus,
  Graphics, StdCtrls, ExtCtrls, DB, DBTables;

type
  TDBCoords = class(TImage)
  private
    FDataLink: TFieldDataLink; { DataLink }
    FReadOnly: boolean;
    FOldOrigin: TPoint;
    FBackGnd: TBitMap;
    FAndImage, FOrImage: TBitMap;
    FPosStr: string;
    function GetDataField: string;
    function GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure DataChange(Sender: TObject);
    procedure UpdateData(Sender: TObject); { called when control changes }
    procedure ChangeData(XPos, YPos: integer);
    procedure ShowLocation(XPos, YPos: integer);
    procedure SetReadOnly(Value: boolean);
    procedure SetBackGnd(ABackGnd: TBitmap);
    procedure SetAndImage(AAndImage: TBitmap);
    procedure SetOrImage(AOrImage: TBitmap);
    property OldOrigin: TPoint read FOldOrigin write FOldOrigin;
    property PosStr: string read FPosStr write FPosStr;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property BackGnd: TBitmap read FBackGnd write SetBackGnd;
    property AndImage: TBitMap read FAndImage write SetAndImage;
    property OrImage: TBitMap read FOrImage write SetOrImage;
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property ReadOnly: boolean read FReadOnly write SetReadOnly;
  end;

procedure Register;

implementation

{ functions for text parsing }
function LeftChop(S: String; len: Byte): String;
begin
  IF length(S) > len THEN
    begin
      MOVE(S[succ(length(S) - len)],
           S[1], Len);
      S[0] := Char(Len);
    end;
  LeftChop := S;
end;

function TrimLead(S: String; C: Char): string;
VAR P : Byte;
begin
  P := 1;
  WHILE (S[P] = C) AND (P <= length(S)) DO Inc(P);
  CASE P OF
    0 : S[0] := #0; {string was 255 of C!}
    1 : ; {not found}
    ELSE
      Move(S[P], S[1], succ(length(S) - P));
      Dec(S[0], pred(P));
  end;
  TrimLead := S;
end;

function Chop(S: String; len: Byte): String;
begin
  IF length(S) > len THEN
    S[0] := Char(Len);
  Chop := S;
end;

function Trim(S: String; C: Char): string;
begin
  WHILE S[length(S)] = C DO Dec(S[0]);
  Trim := S;
end;

function TrimTo(VAR S: String; C: Char): string;
begin
  if Pos(C, S) <> 0 then
  TrimTo := LeftChop(S, Length(S)-Pos(C, S));
end;

function TrimFrom(VAR S: String; C: Char): string;
begin
  if Pos(C, S) <> 0 then
  TrimFrom := Chop(S, Pos(C, S));
end;

{ TDBCoords }
constructor TDBCoords.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBackGnd := TBitMap.Create;  { Create the bitmaps for }
  FAndImage := TBitMap.Create; { displaying the sprites }
  FOrImage := TBitMap.Create;
  FDataLink := TFieldDataLink.Create; { create datalink }
  FDataLink.OnDataChange := DataChange; { attach method to event }
  FDataLink.OnUpdateData := UpdateData; { attach method to event }
end;

destructor TDBCoords.Destroy;
begin
  if Assigned(FBackGnd) then FBackGnd.Destroy;
  if Assigned(FAndImage) then FAndImage.Destroy;
  if Assigned(FOrImage) then FOrImage.Destroy;
  FDataLink.Free;        { free resources }
  FDataLink := nil;
  inherited Destroy;
end;

procedure TDBCoords.SetBackGnd(ABackGnd: TBitmap);
begin
 { Copy data from source bitmap }
 { it would have been nice to just copy
   the primary image but I couldn't find
   an appropriate event to do the assign() }
  FBackGnd.Assign(ABackGnd);
  Invalidate;
end;

procedure TDBCoords.SetAndImage(AAndImage: TBitmap);
begin
 {Copy And image data from source bitmap}
  FAndImage.Assign(AAndImage);
  Invalidate;
end;

procedure TDBCoords.SetOrImage(AOrImage: TBitmap);
begin
 {Copy Or image data from source bitmap}
  FOrImage.Assign(AOrImage);
  Invalidate;
end;

procedure TDBCoords.ShowLocation(XPos, YPos: integer);
var Temp: TPoint;
begin
  { don't show the sprite during design time
  because it stays om the primary image - any
  coords less than zero blank out the sprite }
  if not (csDesigning in ComponentState) and ((XPos >= 0) and (YPos >= 0)) then
  begin
    { erase old sprite }
    BitBlt(Canvas.Handle, OldOrigin.X, OldOrigin.Y, AndImage.Width, AndImage.Height,
           BackGnd.Canvas.Handle, OldOrigin.X, OldOrigin.Y, SrcCopy);
    { AND the AndImage to the canvas }
    BitBlt(Canvas.Handle, XPos, YPos, AndImage.Width, AndImage.Height,
           AndImage.Canvas.Handle, 0, 0, SRCAND);
    { COPY the OrImage to the canvas }
    BitBlt(Canvas.Handle, XPos, YPos, AndImage.Width, AndImage.Height,
           OrImage.Canvas.Handle, 0, 0, SRCPAINT);
  end else
    { just erase the old sprite }
    BitBlt(Canvas.Handle, OldOrigin.X, OldOrigin.Y, AndImage.Width, AndImage.Height,
           BackGnd.Canvas.Handle, OldOrigin.X, OldOrigin.Y, SrcCopy);
  { Redraw image - A windows InvalidateRect might be better here }
  Invalidate;
  { Store location of Image }
  Temp.X := XPos;
  Temp.Y := YPos;
  OldOrigin := Temp;
end;

procedure TDBCoords.DataChange(Sender: TObject);
var
  S,S1,S2: string;
begin
  { Data handling code here }
  if not (FDataLink.Field = nil) then { if there is no field assigned... }
  begin
    if not FDataLink.Field.IsNull then { for null field erase sprite }
    begin
      { Parse field }
      S := FDataLink.Field.AsString;
      if S <> '' then
      begin
        S1 := Trim(TrimFrom(S, ','), ',');
        S2 := TrimLead(TrimTo(S, ','), ',');
        ShowLocation(StrToInt(S1), StrToInt(S2)); { Draw sprite }
      end;
    end else ShowLocation(-1, -1); { erase sprite }
  end else ShowLocation(-1, -1);  { erase sprite }
end;

{ UpdateData is only called after calls to both FDataLink.Modified and
  FDataLink.UpdateRecord. }
procedure TDBCoords.UpdateData(Sender: TObject);
begin
  { set field data to mouse coords }
  FDataLink.Field.AsString := PosStr;
end;

procedure TDBCoords.ChangeData(XPos, YPos: integer);
begin
  PosStr := IntToStr(XPos)+','+IntToStr(YPos); { record mouse location }
  ShowLocation(XPos, YPos);  { draw sprite }
  FDataLink.Modified; { tell data link that data changed }
end;

{ Only process the mouse-down if the data link can edit the data. Otherwise,
  just call the event handler to let the user handle the mouse-down event. }
procedure TDBCoords.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  MyMouseDown: TMouseEvent;
begin
  { This method only leaves Button = mbCenter for user processing
    but since you can't redeclare methods as private the method shows
    up in the object browser }
  if not ReadOnly and FDataLink.Edit then
    if (Button = mbLeft) then ChangeData(X, Y)
    else if (Button = mbRight) then ChangeData(-1, -1)
  else
  begin
    MyMouseDown := OnMouseDown;
    if Assigned(MyMouseDown) then MyMouseDown(Self, Button, Shift, X, Y);
  end;
end;

{ Data source methods }
function TDBCoords.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TDBCoords.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

function TDBCoords.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

procedure TDBCoords.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

procedure TDBCoords.SetReadOnly(Value: boolean);
begin
  FReadOnly := Value;
end;

{ register component }
procedure Register;
begin
  RegisterComponents('Data Controls', [TDBCoords]);
end;

end.
